#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <SDL2/SDL.h>
#include <string.h>
#include <time.h>
#include <stdbool.h>

#include "logger.h"
#include "game.h" 
#include "draw.h" 
#include "garbo.h" 

const int screen_width = 800;
const int screen_height = 600;

struct SDL_Window* make_window(void) {
    if (SDL_Init(SDL_INIT_EVERYTHING) != 0) { 
        printf("error initializing SDL: %s\n", SDL_GetError()); 
    } 


    return SDL_CreateWindow("space game", 
                                       SDL_WINDOWPOS_CENTERED, 
                                       SDL_WINDOWPOS_CENTERED, 
                                       0, 0, 
                                       SDL_WINDOW_FULLSCREEN_DESKTOP); 
}


void redraw(struct SDL_Renderer * ren) {
    // check time
    // redraw if 1/60th of second passed
    
    SDL_RenderClear(ren);
    redraw_buffer(ren);
    SDL_RenderPresent(ren);
}

int physics_loop(void *ptr) {
    while (true) {
        step(10);
        SDL_Delay(2);
    }
}


int game(void) {
    LOGLEVEL = DEBUG;
    STDOUTLEVEL = DEBUG;

//    SDL_SetHint( SDL_HINT_RENDER_SCALE_QUALITY, "2" );

    logwrite(INFO, "Starting\n");
    SDL_Window * win = make_window();
    SDL_Renderer * ren = SDL_CreateRenderer(win, -1, SDL_RENDERER_ACCELERATED
            | SDL_RENDERER_PRESENTVSYNC);

    queue_for_cleanup(win, WINDOW);
    queue_for_cleanup(ren, RENDERER);

  //  IMG_Init(IMG_INIT_PNG | IMG_INIT_JPG);

    if (ren == NULL) {
        SDL_DestroyWindow(win);
        SDL_Quit();
    }

    int close = 0;

    //draw_pictures(ren);
    
    SDL_Thread *physics_thread;
    int ignore;

    bool once = true;
    startgame(ren);

    physics_thread = SDL_CreateThread(physics_loop, "Physics", (void *)NULL);

    while (!close)  {

        SDL_Event event; 
        while(SDL_PollEvent(&event))  {
            switch (event.type) {
                case SDL_QUIT:
                    close = 1;
                    return 0;
                case SDL_KEYDOWN:
                case SDL_KEYUP:
                case SDL_MOUSEBUTTONDOWN:
                case SDL_MOUSEBUTTONUP:
                    handle_input_event (event);
            }
        }
        /* Redraw Screen */
//        step(10);
        redraw(ren);
    }

}

int main (int argc, char** argv)  {

    game();
    SDL_Quit();
    //empty_cleanup_queue();
    return 0; 
} 

